@@ -1,4 +1,5 @@
 /*
+ * SPDX-License-Identifier: BSD-3-Clause
  * Copyright © 2009 CNRS
  * Copyright © 2009-2025 Inria.  All rights reserved.
  * Copyright © 2009-2012, 2020 Université Bordeaux
@@ -187,9 +188,6 @@ static PFN_GETACTIVEPROCESSORGROUPCOUNT GetActiveProcessorGroupCountProc;
 typedef WORD (WINAPI *PFN_GETACTIVEPROCESSORCOUNT)(WORD);
 static PFN_GETACTIVEPROCESSORCOUNT GetActiveProcessorCountProc;
 
-typedef DWORD (WINAPI *PFN_GETCURRENTPROCESSORNUMBER)(void);
-static PFN_GETCURRENTPROCESSORNUMBER GetCurrentProcessorNumberProc;
-
 typedef VOID (WINAPI *PFN_GETCURRENTPROCESSORNUMBEREX)(PPROCESSOR_NUMBER);
 static PFN_GETCURRENTPROCESSORNUMBEREX GetCurrentProcessorNumberExProc;
 
@@ -202,18 +200,9 @@ static PFN_SETTHREADGROUPAFFINITY SetThreadGroupAffinityProc;
 typedef BOOL (WINAPI *PFN_GETTHREADGROUPAFFINITY)(HANDLE hThread, PGROUP_AFFINITY GroupAffinity);
 static PFN_GETTHREADGROUPAFFINITY GetThreadGroupAffinityProc;
 
-typedef BOOL (WINAPI *PFN_GETNUMAAVAILABLEMEMORYNODE)(UCHAR Node, PULONGLONG AvailableBytes);
-static PFN_GETNUMAAVAILABLEMEMORYNODE GetNumaAvailableMemoryNodeProc;
-
 typedef BOOL (WINAPI *PFN_GETNUMAAVAILABLEMEMORYNODEEX)(USHORT Node, PULONGLONG AvailableBytes);
 static PFN_GETNUMAAVAILABLEMEMORYNODEEX GetNumaAvailableMemoryNodeExProc;
 
-typedef LPVOID (WINAPI *PFN_VIRTUALALLOCEXNUMA)(HANDLE hProcess, LPVOID lpAddress, SIZE_T dwSize, DWORD flAllocationType, DWORD flProtect, DWORD nndPreferred);
-static PFN_VIRTUALALLOCEXNUMA VirtualAllocExNumaProc;
-
-typedef BOOL (WINAPI *PFN_VIRTUALFREEEX)(HANDLE hProcess, LPVOID lpAddress, SIZE_T dwSize, DWORD dwFreeType);
-static PFN_VIRTUALFREEEX VirtualFreeExProc;
-
 typedef BOOL (WINAPI *PFN_QUERYWORKINGSETEX)(HANDLE hProcess, PVOID pv, DWORD cb);
 static PFN_QUERYWORKINGSETEX QueryWorkingSetExProc;
 
@@ -224,39 +213,35 @@ static void hwloc_win_get_function_ptrs(void)
 {
   HMODULE kernel32, ntdll;
 
-#if HWLOC_HAVE_GCC_W_CAST_FUNCTION_TYPE
+#if (defined HWLOC_HAVE_GCC_W_CAST_FUNCTION_TYPE) && HWLOC_HAVE_GCC_W_CAST_FUNCTION_TYPE
 #pragma GCC diagnostic ignored "-Wcast-function-type"
 #endif
 
     kernel32 = LoadLibrary(TEXT("kernel32.dll"));
     if (kernel32) {
+      /* Windows7 and Server 2008 R2 */
       GetActiveProcessorGroupCountProc =
 	(PFN_GETACTIVEPROCESSORGROUPCOUNT) GetProcAddress(kernel32, "GetActiveProcessorGroupCount");
       GetActiveProcessorCountProc =
 	(PFN_GETACTIVEPROCESSORCOUNT) GetProcAddress(kernel32, "GetActiveProcessorCount");
-      GetCurrentProcessorNumberProc =
-	(PFN_GETCURRENTPROCESSORNUMBER) GetProcAddress(kernel32, "GetCurrentProcessorNumber");
       GetCurrentProcessorNumberExProc =
 	(PFN_GETCURRENTPROCESSORNUMBEREX) GetProcAddress(kernel32, "GetCurrentProcessorNumberEx");
       SetThreadGroupAffinityProc =
 	(PFN_SETTHREADGROUPAFFINITY) GetProcAddress(kernel32, "SetThreadGroupAffinity");
       GetThreadGroupAffinityProc =
 	(PFN_GETTHREADGROUPAFFINITY) GetProcAddress(kernel32, "GetThreadGroupAffinity");
-      GetNumaAvailableMemoryNodeProc =
-	(PFN_GETNUMAAVAILABLEMEMORYNODE) GetProcAddress(kernel32, "GetNumaAvailableMemoryNode");
       GetNumaAvailableMemoryNodeExProc =
 	(PFN_GETNUMAAVAILABLEMEMORYNODEEX) GetProcAddress(kernel32, "GetNumaAvailableMemoryNodeEx");
       GetLogicalProcessorInformationExProc =
 	(PFN_GETLOGICALPROCESSORINFORMATIONEX)GetProcAddress(kernel32, "GetLogicalProcessorInformationEx");
+      /* Windows7 and Server 2008 R2,
+       * and PSAPI_VERSION not set to 1 in Psapi.h */
       QueryWorkingSetExProc =
 	(PFN_QUERYWORKINGSETEX) GetProcAddress(kernel32, "K32QueryWorkingSetEx");
-      VirtualAllocExNumaProc =
-	(PFN_VIRTUALALLOCEXNUMA) GetProcAddress(kernel32, "VirtualAllocExNuma");
-      VirtualFreeExProc =
-	(PFN_VIRTUALFREEEX) GetProcAddress(kernel32, "VirtualFreeEx");
     }
 
     if (!QueryWorkingSetExProc) {
+      /* Vista and Server 2003 with SP1 and Server 2008 */
       HMODULE psapi = LoadLibrary(TEXT("psapi.dll"));
       if (psapi)
         QueryWorkingSetExProc = (PFN_QUERYWORKINGSETEX) GetProcAddress(psapi, "QueryWorkingSetEx");
@@ -532,16 +517,16 @@ hwloc_windows_get_processor_group_cpuset(hwloc_topology_t topology, unsigned pg_
 static int
 hwloc_win_get_thisthread_last_cpu_location(hwloc_topology_t topology __hwloc_attribute_unused, hwloc_cpuset_t set, int flags __hwloc_attribute_unused)
 {
-  assert(GetCurrentProcessorNumberExProc || (GetCurrentProcessorNumberProc && nr_processor_groups == 1));
+  assert(GetCurrentProcessorNumberExProc || nr_processor_groups == 1);
 
-  if (nr_processor_groups > 1 || !GetCurrentProcessorNumberProc) {
+  if (nr_processor_groups > 1) {
     PROCESSOR_NUMBER num;
     GetCurrentProcessorNumberExProc(&num);
     hwloc_bitmap_from_ith_ULONG_PTR(set, num.Group, ((ULONG_PTR)1) << num.Number);
     return 0;
   }
 
-  hwloc_bitmap_from_ith_ULONG_PTR(set, 0, ((ULONG_PTR)1) << GetCurrentProcessorNumberProc());
+  hwloc_bitmap_from_ith_ULONG_PTR(set, 0, ((ULONG_PTR)1) << GetCurrentProcessorNumber());
   return 0;
 }
 
@@ -834,14 +819,14 @@ hwloc_win_alloc_membind(hwloc_topology_t topology __hwloc_attribute_unused, size
   }
 
   node = hwloc_bitmap_first(nodeset);
-  return VirtualAllocExNumaProc(GetCurrentProcess(), NULL, len, MEM_COMMIT|MEM_RESERVE, PAGE_EXECUTE_READWRITE, node);
+  return VirtualAllocExNuma(GetCurrentProcess(), NULL, len, MEM_COMMIT|MEM_RESERVE, PAGE_EXECUTE_READWRITE, node);
 }
 
 static int
 hwloc_win_free_membind(hwloc_topology_t topology __hwloc_attribute_unused, void *addr, size_t len __hwloc_attribute_unused) {
   if (!addr)
     return 0;
-  if (!VirtualFreeExProc(GetCurrentProcess(), addr, 0, MEM_RELEASE))
+  if (!VirtualFreeEx(GetCurrentProcess(), addr, 0, MEM_RELEASE))
     return -1;
   return 0;
 }
@@ -1189,7 +1174,7 @@ hwloc_look_windows(struct hwloc_backend *backend, struct hwloc_disc_status *dsta
 	      obj->nodeset = hwloc_bitmap_alloc();
 	      hwloc_bitmap_set(obj->nodeset, id);
 	      if ((GetNumaAvailableMemoryNodeExProc && GetNumaAvailableMemoryNodeExProc(id, &avail))
-		  || (GetNumaAvailableMemoryNodeProc && GetNumaAvailableMemoryNodeProc(id, &avail))) {
+		  || GetNumaAvailableMemoryNode(id, &avail)) {
 	        obj->attr->numanode.local_memory = avail;
 		gotnumamemory++;
 	      }
@@ -1356,7 +1341,7 @@ void
 hwloc_set_windows_hooks(struct hwloc_binding_hooks *hooks,
 			struct hwloc_topology_support *support)
 {
-  if (GetCurrentProcessorNumberExProc || (GetCurrentProcessorNumberProc && nr_processor_groups == 1))
+  if (GetCurrentProcessorNumberExProc || nr_processor_groups == 1)
     hooks->get_thisthread_last_cpu_location = hwloc_win_get_thisthread_last_cpu_location;
 
   if (nr_processor_groups == 1) {
@@ -1380,12 +1365,10 @@ hwloc_set_windows_hooks(struct hwloc_binding_hooks *hooks,
     hooks->get_thisthread_membind = hwloc_win_get_thisthread_membind;
   }
 
-  if (VirtualAllocExNumaProc) {
-    hooks->alloc_membind = hwloc_win_alloc_membind;
-    hooks->alloc = hwloc_win_alloc;
-    hooks->free_membind = hwloc_win_free_membind;
-    support->membind->bind_membind = 1;
-  }
+  hooks->alloc_membind = hwloc_win_alloc_membind;
+  hooks->alloc = hwloc_win_alloc;
+  hooks->free_membind = hwloc_win_free_membind;
+  support->membind->bind_membind = 1;
 
   if (QueryWorkingSetExProc && max_numanode_index <= 63 /* PSAPI_WORKING_SET_EX_BLOCK.Node is 6 bits only */)
     hooks->get_area_memlocation = hwloc_win_get_area_memlocation;
