@@ -1,5 +1,6 @@
 /*
- * Copyright © 2010-2024 Inria.  All rights reserved.
+ * SPDX-License-Identifier: BSD-3-Clause
+ * Copyright © 2010-2026 Inria.  All rights reserved.
  * Copyright © 2010-2013 Université Bordeaux
  * Copyright © 2010-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -488,7 +489,7 @@ static void read_amd_cores_legacy(struct procinfo *infos, struct cpuiddump *src_
 }
 
 /* AMD unit/node from CPUID 0x8000001e leaf (topoext) */
-static void read_amd_cores_topoext(struct hwloc_x86_backend_data_s *data, struct procinfo *infos, unsigned long flags, struct cpuiddump *src_cpuiddump)
+static void read_amd_cores_topoext(struct hwloc_x86_backend_data_s *data, struct procinfo *infos, unsigned long flags __hwloc_attribute_unused, struct cpuiddump *src_cpuiddump)
 {
   unsigned apic_id, nodes_per_proc = 0;
   unsigned eax, ebx, ecx, edx;
@@ -497,7 +498,6 @@ static void read_amd_cores_topoext(struct hwloc_x86_backend_data_s *data, struct
   cpuid_or_from_dump(&eax, &ebx, &ecx, &edx, src_cpuiddump);
   infos->apicid = apic_id = eax;
 
-  if (flags & HWLOC_X86_DISC_FLAG_TOPOEXT_NUMANODES) {
     if (infos->cpufamilynumber == 0x16) {
       /* ecx is reserved */
       infos->ids[NODE] = 0;
@@ -512,7 +512,6 @@ static void read_amd_cores_topoext(struct hwloc_x86_backend_data_s *data, struct
         || (infos->cpufamilynumber == 0x19 && nodes_per_proc > 1)) {
       hwloc_debug("warning: undefined nodes_per_proc value %u, assuming it means %u\n", nodes_per_proc, nodes_per_proc);
     }
-  }
 
   if (infos->cpufamilynumber <= 0x16) { /* topoext appeared in 0x15 and compute-units were only used in 0x15 and 0x16 */
     unsigned cores_per_unit;
@@ -655,7 +654,13 @@ static void look_proc(struct hwloc_backend *backend, struct procinfo *infos, uns
   cpuid_or_from_dump(&eax, &ebx, &ecx, &edx, src_cpuiddump);
   infos->apicid = ebx >> 24;
   if (edx & (1 << 28)) {
-    legacy_max_log_proc = 1 << hwloc_flsl(((ebx >> 16) & 0xff) - 1);
+    unsigned ebx_16_23 = (ebx >> 16) & 0xff;
+    if (ebx_16_23) {
+      legacy_max_log_proc = 1 << hwloc_flsl(ebx_16_23 - 1);
+    } else {
+      hwloc_debug("HTT bit set in CPUID 0x01.edx, but legacy_max_proc = 0 in ebx, assuming legacy_max_log_proc = 1\n");
+      legacy_max_log_proc = 1;
+    }
   } else {
     hwloc_debug("HTT bit not set in CPUID 0x01.edx, assuming legacy_max_log_proc = 1\n");
     legacy_max_log_proc = 1;
@@ -1417,20 +1422,46 @@ look_procs(struct hwloc_backend *backend, struct procinfo *infos, unsigned long
       /* use hybrid info for cpukinds */
       if (cpuid_type == intel) {
         /* Hybrid Intel */
+        hwloc_bitmap_t lpset = hwloc_bitmap_alloc();
         hwloc_bitmap_t atomset = hwloc_bitmap_alloc();
         hwloc_bitmap_t coreset = hwloc_bitmap_alloc();
+        unsigned max_cache_levels = 0;
+        int efficiency;
         for(i=0; i<nbprocs; i++) {
-          if (infos[i].hybridcoretype == 0x20)
-            hwloc_bitmap_set(atomset, i);
-          else if (infos[i].hybridcoretype == 0x40)
+          if (infos[i].numcaches > max_cache_levels)
+            max_cache_levels = infos[i].numcaches;
+        }
+        for(i=0; i<nbprocs; i++) {
+          if (infos[i].hybridcoretype == 0x20) {
+            /* On Family 6 hybrids, Atom cores without an L3 cache are low-power cores */
+            if (infos[i].cpufamilynumber == 6 && infos[i].numcaches < max_cache_levels)
+              hwloc_bitmap_set(lpset, i);
+            else
+              hwloc_bitmap_set(atomset, i);
+          } else if (infos[i].hybridcoretype == 0x40) {
             hwloc_bitmap_set(coreset, i);
+          }
+        }
+        /* Lower values indicate less efficient cores. This counter is incremented each time a new
+         * CPU kind is registered, so registration must be done in least-to-most efficient order.
+         */
+        efficiency = 0;
+        /* register IntelLowPower set if any */
+        if (!hwloc_bitmap_iszero(lpset)) {
+          struct hwloc_info_s infoattr;
+          infoattr.name = (char *) "CoreType";
+          infoattr.value = (char *) "IntelLowPower";
+          hwloc_internal_cpukinds_register(topology, lpset, efficiency++, &infoattr, 1, HWLOC_CPUKINDS_REGISTER_FLAG_OVERWRITE_FORCED_EFFICIENCY);
+          /* the cpuset is given to the callee */
+        } else {
+          hwloc_bitmap_free(lpset);
         }
         /* register IntelAtom set if any */
         if (!hwloc_bitmap_iszero(atomset)) {
           struct hwloc_info_s infoattr;
           infoattr.name = (char *) "CoreType";
           infoattr.value = (char *) "IntelAtom";
-          hwloc_internal_cpukinds_register(topology, atomset, HWLOC_CPUKIND_EFFICIENCY_UNKNOWN, &infoattr, 1, 0);
+          hwloc_internal_cpukinds_register(topology, atomset, efficiency++, &infoattr, 1, HWLOC_CPUKINDS_REGISTER_FLAG_OVERWRITE_FORCED_EFFICIENCY);
           /* the cpuset is given to the callee */
         } else {
           hwloc_bitmap_free(atomset);
@@ -1440,7 +1471,7 @@ look_procs(struct hwloc_backend *backend, struct procinfo *infos, unsigned long
           struct hwloc_info_s infoattr;
           infoattr.name = (char *) "CoreType";
           infoattr.value = (char *) "IntelCore";
-          hwloc_internal_cpukinds_register(topology, coreset, HWLOC_CPUKIND_EFFICIENCY_UNKNOWN, &infoattr, 1, 0);
+          hwloc_internal_cpukinds_register(topology, coreset, efficiency++, &infoattr, 1, HWLOC_CPUKINDS_REGISTER_FLAG_OVERWRITE_FORCED_EFFICIENCY);
           /* the cpuset is given to the callee */
         } else {
           hwloc_bitmap_free(coreset);
@@ -1744,7 +1775,7 @@ hwloc_x86_discover(struct hwloc_backend *backend, struct hwloc_disc_status *dsta
 
   if (topology->levels[0][0]->cpuset) {
     /* somebody else discovered things, reconnect levels so that we can look at them */
-    hwloc_topology_reconnect(topology, 0);
+    hwloc__reconnect(topology, 0);
     if (topology->nb_levels == 2 && topology->level_nbobjects[1] == data->nbprocs) {
       /* only PUs were discovered, as much as we would, complete the topology with everything else */
       alreadypus = 1;
@@ -1816,7 +1847,7 @@ hwloc_x86_check_cpuiddump_input(const char *src_cpuiddump_path, hwloc_bitmap_t s
     goto out_with_path;
   }
   fclose(file);
-  if (strcmp(line, "Architecture: x86\n")) {
+  if (strncmp(line, "Architecture: x86", 17)) {
     fprintf(stderr, "hwloc/x86: Found non-x86 dumped cpuid summary in %s: %s\n", path, line);
     goto out_with_path;
   }
