@@ -1,5 +1,6 @@
 /*
- * Copyright © 2017-2019 Inria.  All rights reserved.
+ * SPDX-License-Identifier: BSD-3-Clause
+ * Copyright © 2017-2024 Inria.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -61,12 +62,28 @@ int main(void)
     assert(offset == 648);
 
     offset = offsetof(struct hwloc_topology, first_dist);
-    assert(offset == 696);
+    assert(offset == 704);
     size = sizeof(struct hwloc_internal_distances_s);
     assert(size == 88);
 
+    offset = offsetof(struct hwloc_topology, memattrs);
+    assert(offset == 728);
+    size = sizeof(struct hwloc_internal_memattr_s);
+    assert(size == 32);
+    size = sizeof(struct hwloc_internal_memattr_target_s);
+    assert(size == 48);
+    size = sizeof(struct hwloc_internal_memattr_initiator_s);
+    assert(size == 40);
+    size = sizeof(struct hwloc_internal_location_s);
+    assert(size == 32);
+
+    offset = offsetof(struct hwloc_topology, cpukinds);
+    assert(offset == 744);
+    size = sizeof(struct hwloc_internal_cpukind_s);
+    assert(size == 40);
+
     offset = offsetof(struct hwloc_topology, grouping_next_subkind);
-    assert(offset == 748);
+    assert(offset == 784);
 
     /* fields after this one aren't needed after discovery */
 
@@ -117,22 +134,47 @@ int main(void)
     size = sizeof(struct hwloc_obj);
     assert(size == 248);
     size = sizeof(union hwloc_obj_attr_u);
+#ifdef HWLOC_HAVE_32BITS_PCI_DOMAIN
+    assert(size == 48);
+#else
     assert(size == 40);
+#endif
     size = sizeof(struct hwloc_info_s);
     assert(size == 16);
 
-    size = sizeof(struct hwloc_topology_support);
+#ifndef HWLOC_HAVE_32BITS_PCI_DOMAIN
+    size = sizeof(struct hwloc_pcidev_attr_s);
     assert(size == 24);
+    offset = offsetof(struct hwloc_pcidev_attr_s, class_id);
+    assert(offset == 6);
+    offset = offsetof(struct hwloc_pcidev_attr_s, revision);
+    assert(offset == 16);
+#else
+    size = sizeof(struct hwloc_pcidev_attr_s);
+    assert(size == 24);
+    offset = offsetof(struct hwloc_pcidev_attr_s, class_id);
+    assert(offset == 8);
+    offset = offsetof(struct hwloc_pcidev_attr_s, revision);
+    assert(offset == 18);
+#endif
+
+    size = sizeof(struct hwloc_topology_support);
+    assert(size == 32);
     size = sizeof(struct hwloc_topology_discovery_support);
-    assert(size == 5);
+    assert(size == 6);
     size = sizeof(struct hwloc_topology_cpubind_support);
     assert(size == 11);
     size = sizeof(struct hwloc_topology_membind_support);
-    assert(size == 15);
+    assert(size == 16);
+    size = sizeof(struct hwloc_topology_misc_support);
+    assert(size == 1);
 
     size = sizeof(struct hwloc_distances_s);
     assert(size == 32);
 
+    size = sizeof(struct hwloc_location);
+    assert(size == 16);
+
     size = sizeof(union hwloc_topology_diff_u);
     assert(size == 56);
     size = sizeof(union hwloc_topology_diff_obj_attr_u);
