@@ -1,5 +1,6 @@
 /*
- * Copyright © 2013-2020 Inria.  All rights reserved.
+ * SPDX-License-Identifier: BSD-3-Clause
+ * Copyright © 2013-2021 Inria.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -15,6 +16,7 @@ void usage(const char *callname __hwloc_attribute_unused, FILE *where)
 	fprintf(where, "  --refname <name>  Change the XML reference identifier to <name> in the output\n");
 	fprintf(where, "                    (default is the filename of the first topology\n");
 	fprintf(where, "  --version         Report version and exit\n");
+        fprintf(where, "  -h --help         Show this usage\n");
 }
 
 int main(int argc, char *argv[])
@@ -28,7 +30,12 @@ int main(int argc, char *argv[])
 	unsigned i, j;
 	int err;
 
-	callname = argv[0];
+        callname = strrchr(argv[0], '/');
+        if (!callname)
+          callname = argv[0];
+        else
+          callname++;
+
 	/* skip argv[0], handle options */
 	argc--;
 	argv++;
