@@ -1,5 +1,6 @@
 /*
- * Copyright © 2019-2021 Inria.  All rights reserved.
+ * SPDX-License-Identifier: BSD-3-Clause
+ * Copyright © 2019-2026 Inria.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -14,31 +15,36 @@
 #include "lstopo.h"
 
 extern void JNIbox(int r, int g, int b, int x, int y, int width, int height, unsigned style, int gp_index, char *info);
-extern void JNItext(char *text, int gp_index, int x, int y, int fontsize, int bold);
+extern void JNItext(char *text, int gp_index, int x, int y, int fontsize, int bold, int outside);
 extern void JNIline(unsigned x1, unsigned y1, unsigned x2, unsigned y2);
 extern void JNIprepare(int width, int height, int fontsize);
 
 #define ANDROID_TEXT_WIDTH(length, fontsize) (((length) * (fontsize))/2)
 #define ANDROID_FONTSIZE_SCALE(size) (((size) * 11) / 9)
 
-static void native_android_box(struct lstopo_output *loutput, const struct lstopo_color *lcolor, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned width, unsigned y, unsigned height, hwloc_obj_t obj, unsigned box_id __hwloc_attribute_unused)
+#define GP_INDEX_MULTIPLIER 100 /* there can be multiple boxes per object, so multiple gp_index by 100 and add the box_id to get a unique box ID */
+
+static void native_android_box(struct lstopo_output *loutput, const struct lstopo_color *lcolor, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned width, unsigned y, unsigned height, hwloc_obj_t obj, unsigned box_id)
 {
     unsigned cpukind_style = lstopo_obj_cpukind_style(loutput, obj);
-    int gp_index = -1;
+    int viewid = -1;
     int r = lcolor->r, g = lcolor->g, b = lcolor->b;
-    char * info = malloc(1096);
+    char info[1096] = "";
     const char * sep = " ";
     unsigned style = 0;
 
     if(obj){
-        gp_index = obj->gp_index;
+        /* we could remove this info for factorized objects and bridges,
+         * but it pretty much doesn't appear at all anyway because it's in the very small bow.
+         */
+        viewid = obj->gp_index * GP_INDEX_MULTIPLIER + box_id;
         hwloc_obj_attr_snprintf(info, 1096, obj, sep, 1);
     }
 
     if (cpukind_style)
         style = cpukind_style;
 
-    JNIbox(r, g, b, x, y, width, height, style, gp_index, info);
+    JNIbox(r, g, b, x, y, width, height, style, viewid, info);
     //Creating a usable java string from char * may trigger an UTF-8 error
     //This code creates a byte array from the char * variable before creating the java string
 }
@@ -62,16 +68,26 @@ static void
 native_android_text(struct lstopo_output *loutput, const struct lstopo_color *lcolor __hwloc_attribute_unused, int size __hwloc_attribute_unused, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned y, const char *text, hwloc_obj_t obj, unsigned text_id __hwloc_attribute_unused)
 {
     unsigned cpukind_style = lstopo_obj_cpukind_style(loutput, obj);
-    int gp_index = -1;
+    int viewid = -1;
     int bold = 0;
+    int outside = 0;
 
-    if(obj)
-        gp_index = obj->gp_index;
+    if(obj) {
+      struct lstopo_obj_userdata *lud = obj->userdata;
+      viewid = obj->gp_index * GP_INDEX_MULTIPLIER;
+      /* no info box in small boxes */
+      if (loutput->factorize_enabled
+          && lud->factorized == 1
+          && obj->parent->arity > loutput->factorize_min[obj->type])
+        outside = 1;
+      else if (obj->type == HWLOC_OBJ_BRIDGE)
+        outside = 1;
+    }
 
     if (cpukind_style % 2)
         bold = 1;
 
-    JNItext((char *)text, gp_index, x, y, loutput->fontsize, bold);
+    JNItext((char *)text, viewid, x, y, loutput->fontsize, bold, outside);
 }
 
 static struct draw_methods native_android_draw_methods = {
